package ru.dbotthepony.kstarbound.client.gl.vertex

import com.google.common.collect.ImmutableList
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.VertexArrayObject


/**
 * Хранит список аттрибутов для применения к Vertex Array Object
 *
 * Аттрибуты плотно упакованы и идут один за другим
 *
 * Создаётся через [GLAttributeList.Builder]
 */
class GLAttributeList(builder: Builder) {
	data class Attribute(val name: String, val index: Int, val glType: GLType, val stride: Int, val offset: Long)

	val attributes: List<Attribute>
	val size get() = attributes.size

	/**
	 * Шаг данных аттрибутов, в байтах. Т.е. одна полная вершина будет занимать [stride] байт в памяти.
	 */
	val stride: Int

	operator fun get(index: Int) = attributes[index]

	init {
		val buildList = ArrayList<Attribute>()

		var offset = 0L
		var stride = 0

		for (i in builder.attributes) {
			stride += i.second.byteSize
		}

		this.stride = stride

		for (i in builder.attributes.indices) {
			val value = builder.attributes[i].second
			buildList.add(Attribute(builder.attributes[i].first, i, value, stride, offset))
			offset += value.byteSize
		}

		attributes = ImmutableList.copyOf(buildList)
	}

	/**
	 * Применяет список атрибутов к заданному [VertexArrayObject] (попутно включая или отключая их через [enable])
	 */
	fun apply(target: VertexArrayObject, enable: Boolean) {
		for (i in attributes.indices) {
			val value = attributes[i]
			target.attribute(i, value.glType.logicalSize, value.glType.typeIndentity, false, value.stride, value.offset)

			if (enable) {
				target.enableAttribute(i)
			}
		}
	}

	class Builder {
		val attributes = ArrayList<Pair<String, GLType>>()

		fun push(type: GLType): Builder {
			return push("$type#${attributes.size}", type)
		}

		fun push(vararg types: GLType): Builder {
			for (type in types) push(type)
			return this
		}

		fun push(name: String, type: GLType): Builder {
			check(!attributes.any { it.first == name }) { "Already has named attribute $name!" }
			attributes.add(name to type)
			return this
		}

		fun build() = GLAttributeList(this)
	}

	companion object {
		val VEC2F = Builder().push(GLType.VEC2F).build()
		val VEC3F = Builder().push(GLType.VEC3F).build()

		val VERTEX_TEXTURE = Builder().push(GLType.VEC3F).push(GLType.VEC2F).build()

		val VERTEX_2D_TEXTURE = Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()

		val TILE = Builder().push(GLType.VEC3F, GLType.VEC2F, GLType.FLOAT).build()
	}
}
